/*
 * FSPointer.java
 * Transform
 * 
 * Copyright (c) 2001-2008 Flagstone Software Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *  * Neither the name of Flagstone Software Ltd. nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.flagstone.transform;

/**
 * FSPointer defines an unsigned 32-bit pointer.
 * 
 * <p>
 * It is thought that this class was used by Macromedia's Generator product and
 * its exact function is not known. It is included in the framework for
 * completeness.
 * </p>
 * 
 * <table class="datasheet">
 * 
 * <tr>
 * <th align="left" colspan="2">Attributes</th>
 * </tr>
 * 
 * <tr>
 * <td>type</td>
 * <td>Identifies the data structure when it is encoded. Read-only.</td>
 * </tr>
 * 
 * <tr>
 * <td>pointer</td>
 * <td>The pointer value.</td>
 * </tr>
 * </table>
 * 
 * <h1 class="datasheet">History</h1>
 * 
 * <p>
 * The FSPointer class represents the DefineBitsPtr tag from the Macromedia
 * Flash (SWF) File Format Specification. It is not documented in later editions
 * of the specification. It is not known when this tag was added.
 * </p>
 */
public class FSPointer extends FSMovieObject
{
	private int pointer = 0;

	/**
	 * Construct an FSPointer object, initialising it with values decoded from an
	 * encoded object.
	 * 
	 * @param coder
	 *            an FSCoder containing the binary data.
	 */
	public FSPointer(FSCoder coder)
	{
		super(DefineBitsPtr);
		decode(coder);
	}

	/**
	 * Constructs an FSPointer object, specifying the pointer.
	 * 
	 * @param aPointer
	 *            the pointer.
	 */
	public FSPointer(int aPointer)
	{
		super(DefineBitsPtr);
		setPointer(aPointer);
	}

	/**
	 * Constructs an FSPointer object by copying values from an existing object.
	 * 
	 * @param obj
	 *            an FSPointer object.
	 */
	public FSPointer(FSPointer obj)
	{
		super(obj);
		pointer = obj.pointer;
	}

	/**
	 * Gets the value for the pointer.
	 * 
	 * @return the pointer value.
	 */
	public int getPointer()
	{
		return pointer;
	}

	/**
	 * Sets the pointer.
	 * 
	 * @param aPointer
	 *            the pointer.
	 */
	public void setPointer(int aPointer)
	{
		pointer = aPointer;
	}

	public boolean equals(Object anObject)
	{
		boolean result = false;

		if (super.equals(anObject))
		{
			FSPointer typedObject = (FSPointer) anObject;

			result = pointer == typedObject.pointer;
		}
		return result;
	}

	public void appendDescription(StringBuffer buffer, int depth)
	{
		buffer.append(name());

		if (depth > 0)
		{
			buffer.append(": { ");
			Transform.append(buffer, "pointer", pointer);
			buffer.append("}");
		}
	}

	public int length(FSCoder coder)
	{
		super.length(coder);
		length += 4;

		return length;
	}

	public void encode(FSCoder coder)
	{
		super.encode(coder);
		coder.writeWord(pointer, 4);
		coder.endObject(name());
	}

	public void decode(FSCoder coder)
	{
		super.decode(coder);
		pointer = coder.readWord(4, false);
		coder.endObject(name());
	}
}
